import pytest

from unq.services.rate_limiter import RateLimiter


@pytest.fixture(autouse=True)
def default_rate_limiter():
    rate_limiter = RateLimiter(repetition_interval=0.5)
    yield rate_limiter
    rate_limiter.stop()


from typing import Any, Callable, Literal, Protocol, overload

class _AbstractSubmittable(Protocol):
    @overload
    def submit(self, function: Callable, keep_result: Literal[True], *args: Any, **kwargs: Any) -> Any: ...
    @overload
    def submit(self, function: Callable, keep_result: Literal[False] = False, *args: Any, **kwargs: Any) -> None:
        """Submit a function to execute later.

        Args:
            callable (Callable): The function to execute.
            keep_results (bool): If this is set to `False`, no `Future` object will be created
            and the result of the function execution will not be recorded.
            It is advised to keep this setting `False` when running in a non-async context. Defaults to `False`.
            args: Any non-keyword arguments to pass to the function.
            kwargs: Any keyword arguments to pass to the function.

        Returns:
            Future | None: A future to the function call result or `None` if keep_result is `False`
        """


from concurrent.futures import ThreadPoolExecutor, Future
from functools import wraps
from queue import Queue
from threading import Condition, Lock, Thread
from time import sleep
from typing import Any, Callable, Generic, Literal, overload, TypeVar

from unq.models import RepetitionInterval, _FutureFunctionCall
from unq.services._context_managed_rate_limiter import _ContextManagedRateLimiter
from unq.services.abstract._abstract_rate_limiter import _AbstractRateLimiter

_RepetitionIntervalType = RepetitionInterval | float
_T = TypeVar("_T")

class RateLimiter(_AbstractRateLimiter):
    """
    `RateLimiter` class, the core service of Unq.
    Allows calling functions at a set rate, unloading functions at the order they were pushed in.

    This calss is concurrent; It manages threads in the background, and it's important to 
    mention that the main thread is non-daemonic so you will need to call `RateLimiter.stop()` in order to end
    the program execution properly.
    
    It is possible to receive the results of the functions by calling `.result()` on the `Future` objects,
    which are `concurrent.futures.Future` objects, blocking until a result is received.

    Args:
        repetition_interval (_RepetitionIntervalType): The interval between each repetition.
        Can either be set through the library `RepetitionInterval` datatype or with a simple
        `float` or a type that is explicitly castable to `float` (has a `__float__` method)
        representing the amount of seconds to sleep between each repetition.

    Raises:
        TypeError: Raised then the value is of an unsupported type
    """

    def __init__(self, repetition_interval: _RepetitionIntervalType) -> None:
        self._rep_interval_lock = Lock()
        self._typecheck_set_repetition_interval_type(repetition_interval)

        self._stop_lock = Lock()
        self._stopped: bool = True

        self._queue_lock = Lock()
        self._queue_condition = Condition(Lock())
        self._call_queue: Queue[_FutureFunctionCall] = Queue()

        self._internal_runner_thread = Thread(None, self._run)  # To be overwritten
        self._executor = ThreadPoolExecutor()

    @overload
    def submit(self, function: Callable[..., _T], keep_result: Literal[True], *args: Any, **kwargs: Any) -> Future[_T]: ...
    @overload
    def submit(self, function: Callable[..., _T], keep_result: Literal[False] = False, *args: Any, **kwargs: Any) -> None: ...
    def submit(self, function: Callable[..., _T], keep_result: bool = False, *args: Any, **kwargs: Any) -> Future[_T] | None:
        if keep_result:
            future: Future[_T] | None = Future()
        else:
            future = None

        function_call = _FutureFunctionCall(future, function, args, kwargs)
        with self._queue_lock:
            self._call_queue.put(function_call)
            self._notify_queue_condition()
        return future

    def start(self) -> None:
        if self.stopped:
            with self._stop_lock:
                self._internal_runner_thread = Thread(None, self._run)
                self._internal_runner_thread.start()
                self._stopped = False

    def stop(self) -> None:
        if not self.stopped:
            with self._stop_lock:
                self._stopped = True
                self._notify_queue_condition()
            self._internal_runner_thread.join()
            
    @overload
    def limit(self, keep_result: Literal[True]) -> Callable[[Callable[..., _T]], Callable[..., Future[_T]]]: ...
    @overload
    def limit(self, keep_result: Literal[False] = False) -> Callable[[Callable[..., _T]], Callable[..., None]]: ...
    def limit(self, keep_result: Literal[True] | Literal[False] = False):
        """Decorator that wraps a function in the rate limiter, will submit the function automatically
        when it's called instead of just calling the function normally.

        Args:
            keep_result (bool, optional): If set to `True`, each function call will return a Future of its return type. Defaults to False.
        """
        def ext_wrapper(func: Callable[..., _T]):
            @wraps(func)
            def wrapper(*args, **kwargs) -> None | Future[_T]:
                return self.submit(func, keep_result, *args, **kwargs)

            return wrapper
        
        return ext_wrapper

    @property
    def stopped(self) -> bool:
        """Whether the rate limiter is stopped."""
        with self._stop_lock:
            return self._stopped

    @property
    def repetition_interval(self) -> float:
        """Repetition interval in seconds

        Returns:
            float: The number of seconds of the repetition interval
        """
        with self._rep_interval_lock:
            return self._repetition_interval

    @repetition_interval.setter
    def repetition_interval(self, repetition_interval: _RepetitionIntervalType):
        """Set the repetition interval while type checking the type and converting the value if needed

        Args:
            repetition_interval (_RepetitionIntervalType): The interval between each repetition.
            Can either be set through the library `RepetitionInterval` datatype or with a simple
            `float` (or castable) representing the amount of seconds to sleep between each repetition.

        Raises:
            TypeError: Raised then the value is of an unsupported type
        """
        with self._rep_interval_lock:
            self._typecheck_set_repetition_interval_type(repetition_interval)

    def _run(self) -> None:
        """Internal run function.

        Contains the main loop of the class. Ran in the background by `RateLimiter.run()`
        """
        while not self.stopped:
            with self._queue_condition:
                while self._queue_is_empty() and not self.stopped:
                    self._queue_condition.wait()
                if self.stopped:
                    break
            with self._queue_lock:
                function_call = self._call_queue.get()
            self._executor.submit(self._execute_function_call, function_call)
            sleep(self.repetition_interval)

    def _notify_queue_condition(self):
        """Notify the queue lock condition while acquiring its lock."""
        with self._queue_condition:
            self._queue_condition.notify_all()

    def _queue_is_empty(self):
        with self._queue_lock:
            return self._call_queue.empty()

    def _typecheck_set_repetition_interval_type(self, repetition_interval: _RepetitionIntervalType):
        """Internal implementation of the type checked set for `repetition_interval`

        Args:
            repetition_interval (_RepetitionIntervalType): The interval between each repetition.
            Can either be set through the library `RepetitionInterval` datatype or with a simple
            `float` or a type that is explicitly castable to `float` (has a `__float__` method)
            representing the amount of seconds to sleep between each repetition.

        Raises:
            TypeError: Raised then the value is of an unsupported type
        """
        if isinstance(repetition_interval, float):
            self._repetition_interval = repetition_interval
        elif isinstance(repetition_interval, RepetitionInterval):
            self._repetition_interval = self._calculate_repetition_interval(repetition_interval)
        elif hasattr(repetition_interval, "__float__"):
            self._repetition_interval = float(repetition_interval)
        else:
            raise TypeError("`repetition_interval` was of incorrect type.")

    def _calculate_repetition_interval(self, repetition_interval: RepetitionInterval) -> float:
        """Calculates the repetition interval using data from a `RepetitionInterval` object

        Args:
            repetition_interval (RepetitionInterval): Object that represents a time interval
            limiting function calls.

        Returns:
            float: The calculated interval

        Raises:
            ValueError: Raised in the case that an invalid timeframe value is provided
        """
        timeframe = repetition_interval.timeframe
        base_interval = 1.0
        match timeframe:
            case "seconds" | "s":
                base_interval *= 1
            case "minutes" | "m":
                base_interval *= 60
            case "hours" | "h":
                base_interval *= 3600
            case _:
                raise ValueError("Invalid timeframe set.")

        base_interval /= repetition_interval.times
        base_interval *= repetition_interval.every
        return base_interval

    def _execute_function_call(self, function_call: _FutureFunctionCall) -> None:
        future = function_call.future

        call_partial = lambda: function_call.function(*function_call.args, **function_call.kwargs)
        try:
            result = call_partial()

            if future:
                future.set_result(result)    

        except Exception as error:  # pylint: disable=broad-exception-caught
            if future:
                future.set_exception(error)
    
    def __enter__(self) -> _ContextManagedRateLimiter:
        if not self.stopped:
            raise RuntimeError("Can't use a started RateLimiter as a context manager.")
        
        self.start()
        return _ContextManagedRateLimiter(self)
            
    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.stop()


from dataclasses import dataclass
from typing import Literal


@dataclass(frozen=True)
class RepetitionInterval:
    """
    A model to represent action repetition intervals.

    `timeframe` is the base time interval - seconds, minutes or hours. Defaults to `second`
    `every` is the number of timeframes for each repetition. Defaults to 1
    `times` is the amount of times an action will be repeated during the set interval. Defaults to 1

    The model is validated! If incorrect values are provided a ValueError will be raised.
    """

    timeframe: Literal["seconds", "minutes", "hours", "s", "m", "h"] = "seconds"
    every: int = 1
    times: int = 1



from asyncio import Future
from typing import Any, Callable, Literal, overload
from unq.services.abstract._abstract_submittable import _AbstractSubmittable


class _ContextManagedRateLimiter(_AbstractSubmittable):
    def __init__(self, rate_limiter: _AbstractSubmittable):
        self._rate_limiter = rate_limiter
        
    @overload
    def submit(self, function: Callable, keep_result: Literal[True], *args: Any, **kwargs: Any) -> Future[Any]: ...
    @overload
    def submit(self, function: Callable, keep_result: Literal[False] = False, *args: Any, **kwargs: Any) -> None: ...
    def submit(self, function: Callable, keep_result: Literal[False] | Literal[True] = False, *args: Any, **kwargs: Any) -> Future[Any] | None:
        return self._rate_limiter.submit(function, keep_result, *args, **kwargs)            


from asyncio import new_event_loop
from dataclasses import FrozenInstanceError

import pytest

from unq.models import _FutureFunctionCall


@pytest.fixture
def default_future():
    return new_event_loop().create_future()


@pytest.fixture
def default_ffc(default_future):
    # Arrange
    future = default_future
    function = print
    args = ("Hello, World!",)
    kwargs = {"test": "arg"}
    return _FutureFunctionCall(
        future=future, function=function, args=args, kwargs=kwargs
    )


def test_constructor_properly_creates_object(default_ffc, default_future):
    # Arrange
    future = default_future
    function = print
    args = ("Hello, World!",)
    kwargs = {"test": "arg"}

    # Assert
    assert default_ffc.future == future, "Future did not match"
    assert default_ffc.function == function, "Function did not match"
    assert default_ffc.args == args, "Args did not match"
    assert default_ffc.kwargs == kwargs, "Kwargs did not match"


def test_object_is_immutable_dataclass(default_ffc):
    # Act, Assert
    with pytest.raises(FrozenInstanceError):
        default_ffc.kwargs = {"arg": "value"}



from dataclasses import FrozenInstanceError

import pytest

from unq import RepetitionInterval


@pytest.fixture
def default_ri():
    # Arrange
    timeframe = "second"
    every = 6
    times = 6
    return RepetitionInterval(timeframe=timeframe, every=every, times=times)


def test_constructor_properly_creates_object(default_ri):
    # Arrange
    timeframe = "second"
    every = 6
    times = 6

    # Assert
    assert default_ri.timeframe == timeframe, "Timeframe did not match"
    assert default_ri.every == every, "Every did not match"
    assert default_ri.times == times, "Times did not match"


def test_object_is_immutable_dataclass(default_ri):
    # Act, Assert
    with pytest.raises(FrozenInstanceError):
        default_ri.timeframe = "hour"



from concurrent.futures import Future
import time
from datetime import datetime
from unittest.mock import AsyncMock, Mock

import pytest

from unq import RateLimiter
from unq.models.repetition_interval import RepetitionInterval

from .fixtures import default_rate_limiter, default_rate_limiter


def test_constructor_parses_repetition_interval():
    # Arrange
    rep_interval = RepetitionInterval(timeframe="hours", every=1, times=3)
    expected_rep_interval = 1200  # Three times every hour is every 1200 seconds

    # Act
    rate_limiter = RateLimiter(rep_interval)

    # Assert
    assert (
        rate_limiter.repetition_interval == expected_rep_interval
    ), "Repetition interval was incorrectly calculated"


def test_constructor_accepts_float_as_repetition_interval():
    # Arrange
    rep_interval = 6

    # Act
    rate_limiter = RateLimiter(repetition_interval=rep_interval)

    # Assert
    assert rate_limiter.repetition_interval == rep_interval


def test_constructor_accepts_float_castables_as_repetition_intervals():
    # Arrange
    rep_interval = 6

    # Act
    rate_limiter = RateLimiter(repetition_interval=rep_interval)

    # Assert
    assert rate_limiter.repetition_interval == 6.0


def test_constructor_raises_when_provided_unparsable_type():
    # Arrange
    rep_interval = {}

    # Act, Assert
    with pytest.raises(TypeError):
        RateLimiter(repetition_interval=rep_interval)


def test_marked_stopped_when_constructed(default_rate_limiter: RateLimiter):
    # Assert
    assert default_rate_limiter.stopped == True, "Somehow, it was started"


def test_marked_stopped_when_stopped_without_being_started(
    default_rate_limiter: RateLimiter,
):
    # Act
    default_rate_limiter.stop()

    # Assert
    assert default_rate_limiter.stopped == True, "Somehow, it was started"


def test_marked_not_stopped_when_started(default_rate_limiter: RateLimiter):
    # Act
    default_rate_limiter.start()

    # Assert
    assert default_rate_limiter.stopped == False, "Somehow, it was stopped"


def test_marked_stopped_when_stopped_after_started(default_rate_limiter: RateLimiter):
    # Act
    default_rate_limiter.start()
    default_rate_limiter.stop()

    # Assert
    assert default_rate_limiter.stopped == True, "Somehow, it was not stopped"


def test_function_executed_after_submit_when_started(default_rate_limiter: RateLimiter):
    # Arrange
    default_rate_limiter.start()
    mock_function = Mock(return_value="Test Test")

    # Act
    default_rate_limiter.submit(mock_function)
    time.sleep(2)  # We don't want to undershoot

    # Assert
    mock_function.assert_called_once()


def test_function_executed_after_submit_in_async_context_when_started(
    default_rate_limiter: RateLimiter,
):
    # Arrange
    default_rate_limiter.start()
    mock_function = Mock(return_value="Test Test")

    # Act
    default_rate_limiter.submit(mock_function, True).result()

    # Assert
    mock_function.assert_called_once()
    
    
def test_function_not_executed_when_stopped(default_rate_limiter: RateLimiter):
    # Arrange
    mock_function = Mock(return_value=6)
    default_rate_limiter.start()
    default_rate_limiter.stop()
    
    # Act
    default_rate_limiter.submit(mock_function)
    time.sleep(2)
    
    # Assert
    mock_function.assert_not_called()
    
    
def test_functions_executed_at_correct_intervals(
    default_rate_limiter: RateLimiter,
):
    # Arrange
    MAX_DELAY_INACCURACY_MICROSECONDS = 2000
    default_rate_limiter.start()

    # Act
    time_1: Future[datetime] = default_rate_limiter.submit(datetime.now, True)
    time_2: Future[datetime] = default_rate_limiter.submit(datetime.now, True)
    time_3: Future[datetime] = default_rate_limiter.submit(datetime.now, True)

    delay_1 = time_2.result() - time_1.result()
    delay_2 = time_3.result() - time_2.result()

    delay_delta_1 = delay_1 - delay_2
    delay_delta_2 = delay_2 - delay_1

    # Assert
    assert (
        delay_delta_1.microseconds < MAX_DELAY_INACCURACY_MICROSECONDS
        or delay_delta_2.microseconds < MAX_DELAY_INACCURACY_MICROSECONDS
    )


def test_functions_executed_at_correct_intervals_after_interval_changed(
    default_rate_limiter: RateLimiter,
):
    # Act, Assert
    test_functions_executed_at_correct_intervals(default_rate_limiter)
    default_rate_limiter.repetition_interval = 1
    test_functions_executed_at_correct_intervals(default_rate_limiter)
    

def test_function_executed_with_arguments(default_rate_limiter: RateLimiter):
    # Arrange
    mock_function = Mock(return_value=6)
    default_rate_limiter.start()
    
    # Act
    default_rate_limiter.submit(mock_function, False, 1, kwarg=2)
    time.sleep(2)
    
    # Assert
    mock_function.assert_called_once_with(1, kwarg=2)
    

def test_exception_in_submitted_function_handled(default_rate_limiter: RateLimiter):
    # Arrange
    def raise_exception():
        raise Exception("test")
    
    mock_raises = Mock(side_effect=raise_exception)
    default_rate_limiter.start()
    
    # Act
    default_rate_limiter.submit(mock_raises)
    time.sleep(2)
    
    # Assert
    mock_raises.assert_called_once()
    

def test_context_managed_instance_starts(default_rate_limiter: RateLimiter):
    # Arrange
    mock_function = Mock(return_value=6)
    
    # Act
    with default_rate_limiter as rl:
        print("test")
        rl.submit(mock_function)
        time.sleep(2) # No overshoots
    
    # Assert
    mock_function.assert_called_once()
    

def test_limit_decorator_works(default_rate_limiter: RateLimiter):
    # Arrange
    @default_rate_limiter.limit(True)
    def mock_function():
        return 6
    
    default_rate_limiter.start()
    
    # Act
    result = mock_function()
    
    # Assert
    assert result.result() == 6

    



from asyncio import Future as AsyncFuture
from concurrent.futures import Future as ConcFuture
from dataclasses import dataclass
from typing import Any, Callable


@dataclass(frozen=True)
class _FutureFunctionCall:
    future: AsyncFuture[Any] | ConcFuture[Any] | None
    function: Callable
    args: tuple
    kwargs: dict[str, Any]



from typing import Protocol
from unq.services.abstract._abstract_submittable import _AbstractSubmittable
from unq.services._context_managed_rate_limiter import _ContextManagedRateLimiter

class _AbstractRateLimiter(_AbstractSubmittable, Protocol):
    def start(self): """Start the rate limiter execution."""
    def stop(self): """Stop the rate limiter execution."""
    def __enter__(self) -> _ContextManagedRateLimiter: ...
    def __exit__(self, exception_type, exception_value, exception_traceback): ...